@@ -1,4 +1,4 @@
-/* xchainkeys 0.1 -- exclusive keychains for X11
+/* xchainkeys -- chained keybindings for X11
  * Copyright (C) 2010 Henning Bekel <h.bekel@googlemail.com>
  *
  * This program is free software; you can redistribute it and/or
@@ -51,10 +51,24 @@ XChainKeys_t* xc_new(Display *display) {
   self->debug = False;
   self->timeout = 3000;
   self->delay = 1000;
-  self->root = binding_new();
-  self->root->action = ":root";
+  self->reentry = NULL;
   self->xmodmap = XGetModifierMapping(self->display);
   
+  self->action_names[0] = ":none";
+  self->action_names[1] = ":enter";
+  self->action_names[2] = ":escape";
+  self->action_names[3] = ":abort";
+  self->action_names[4] = ":exec";
+  self->action_names[5] = ":repeat";
+  self->action_names[6] = ":send";
+  self->num_actions = 7;
+
+  self->root = binding_new();
+  self->root->action = XC_ACTION_NONE;
+  
+  /* prepare modmask array, containing all possible combinations of
+   * num, caps and scroll lock */
+
   i = 0;
   num = 
     xc_keycode_to_modmask(self, XKeysymToKeycode(self->display, XStringToKeysym("Num_Lock")));
@@ -72,6 +86,8 @@ XChainKeys_t* xc_new(Display *display) {
   self->modmask[i++] = caps | scroll;
   self->modmask[i++] = num | caps | scroll;
 
+  xc_find_config(self);
+
   return self;
 }
 
@@ -86,13 +102,29 @@ void xc_usage(XChainKeys_t *self) {
   printf("  -h, --help    : Print this help text\n");
   printf("  -v, --version : Print version information\n");
   printf("  -k, --keys    : Show valid keyspecs\n");
+  printf("  -f, --file    : alternative config file\n");
   printf("\n");
 }
 
+void xc_find_config(XChainKeys_t *self) {
+  
+  self->config = (char *) calloc(4096, sizeof(char));
+
+  /* determine config file self->config */
+  strcpy(self->config, getenv("HOME"));
+  
+  if(getenv("XDG_CONFIG_HOME") != NULL ) {
+    strcpy(self->config, getenv("XDG_CONFIG_HOME"));
+    strcat(self->config, "/xchainkeys/xchainkeys.conf");
+  } 
+  else {
+    strcat(self->config, "/.config/xchainkeys/xchainkeys.conf");
+  }
+}
+
 void xc_parse_config(XChainKeys_t *self) {
 
-  FILE *config;
-  char *path = (char *) calloc(4096, sizeof(char));
+  FILE *f;
   char *buffer = (char *) calloc(4096, sizeof(char));
   char *argument= (char *) calloc(4096, sizeof(char));
   char *line, *token, *expect;
@@ -103,33 +135,31 @@ void xc_parse_config(XChainKeys_t *self) {
   Key_t *key;
   Binding_t *binding;
   Binding_t *parent;
-  char *font = "fixed";
-  char *fg = "black";
-  char *bg = "white";
 
-  /* determine config file path */
-  strcpy(path, getenv("HOME"));
-  
-  if(getenv("XDG_CONFIG_HOME") != NULL ) {
-    strcpy(path, getenv("XDG_CONFIG_HOME"));
-    strcat(path, "/xchainkeys/xchainkeys.conf");
-  } 
-  else {
-    strcat(path, "/.config/xchainkeys/xchainkeys.conf");
-  }
+  int feedback = True;
+
+  char *font = calloc(512, sizeof(char));
+  strcpy(font, "fixed");
+
+  char *fg = calloc(64, sizeof(char));
+  strcpy(fg, "black");
+
+  char *bg = calloc(64, sizeof(char));
+  strcpy(bg, "white");
 
   /* try to open config file */
-  config = fopen(path, "r");
+  f = fopen(self->config, "r");
   
-  if(config == NULL) {
-    fprintf(stderr, "error: '%s': %s\n", path, strerror(errno));
+  if(f == NULL) {
+    fprintf(stderr, "%s: error: '%s': %s\n", 
+	    PACKAGE_NAME, self->config, strerror(errno));
     exit(EXIT_FAILURE);
   }
 
-  if (self->debug) fprintf(stderr, "Parsing config file %s...\n", path);
+  if (self->debug) printf("Parsing config file %s\n", self->config);
 
   /* parse file */
-  while(fgets(buffer, 4096, config) != NULL) {
+  while(fgets(buffer, 4096, f) != NULL) {
     linenum++;
     line = buffer;
 
@@ -144,8 +174,6 @@ void xc_parse_config(XChainKeys_t *self) {
     if( line[strlen(line)-1] == '\n' )
       line[strlen(line)-1] = '\0';
     
-    /* if (self->debug) fprintf(stderr, "%s\n", line); */
-
     /* parse options */
     if( strncmp(line, "timeout", 7) == 0 ) {
       line += 7;
@@ -163,6 +191,19 @@ void xc_parse_config(XChainKeys_t *self) {
       continue;
     }
 
+    if( strncmp(line, "feedback", 8) == 0 ) {
+      line += 8;
+      line += strspn(line, ws);
+      line[strcspn(line, ws)] = '\0';
+
+      if(strcmp(line, "off") == 0)
+	feedback = False;
+      if(strcmp(line, "on") == 0) 
+        feedback = True;
+
+      continue;
+    }
+
     if( strncmp(line, "font", 4) == 0 ) {
       line += 4;
       line += strspn(line, ws);
@@ -195,10 +236,10 @@ void xc_parse_config(XChainKeys_t *self) {
     expect = "key";
     argument[0] = '\0';
 
-    while(line[0] != '\0') {
+    while(strlen(line)) {
       len = strcspn(line, ws);
       
-      token = calloc(1, (len+1)*sizeof(char));
+      token = calloc((len+1), sizeof(char));
       strncpy(token, line, len);
 
       line += strcspn(line, ws);
@@ -207,42 +248,38 @@ void xc_parse_config(XChainKeys_t *self) {
       if( token[0] == ':' )
 	expect = "action";
 
-      /* if (self->debug) fprintf(stderr, "%d: (expect %s) '%s'\n", pos, expect, token); */
-
       if (strcmp(expect, "key") == 0) {
 	if ((key = key_new(token)) != NULL) {
 
-	  // if this key is already bound in the parent binding,
-	  // then make that binding the parent binding for the next key
+	  /* if this key is already bound in the parent binding,
+	   * then make that binding the parent binding for the next key
+	   */
 	  if(binding_get_child_by_key(parent, key) != NULL) {
 	    parent = binding_get_child_by_key(parent, key);
 	    goto next_token;
 	  }  
 
-	  // create a binding for this key
+	  /* create a binding for this key */
 	  binding = binding_new();
 	  binding->key = key;
 
-	  // append the new binding to the current parent
+	  /* append the new binding to the current parent */
 	  binding_append_child(parent, binding);
 
-	  if(parent == xc->root && !strlen(binding->action))
-	    binding->action = ":enter";
-
-	  // make this binding the parent for the next
+	  /* make this binding the parent for the next */
 	  parent = binding;
 	}
 	else {
-	  fprintf(stderr, 
-		  "Warning: Invalid keyspec: '%s'. Use %s -k to show valid keyspecs.\n",
-		  token, PACKAGE_NAME);
+	  fprintf(stderr,
+		  "%s: warning: Invalid keyspec: '%s'. Use %s -k to show valid keyspecs.\n",
+		  PACKAGE_NAME, token, PACKAGE_NAME);
 	  fprintf(stderr, "-> Skipping line %d: '%s'\n", linenum, buffer);
 	  goto next_line;
 	}
       }
       
       if (strcmp(expect, "action") == 0) {
-	binding->action = strdup(token);
+	binding_set_action(binding, token);
 	expect = "argument";
 	goto next_token;
       }
@@ -251,38 +288,46 @@ void xc_parse_config(XChainKeys_t *self) {
 	if(strlen(argument))
 	  strncat(argument, " ", 1);
 	strncat(argument, token, strlen(token));
+	goto next_token;
       }
 
     next_token:
       pos++;
       free(token);
     }
-    // done with this line
+    /* all tokens parsed */
     
     if (binding != NULL) {
-      // append the argument to the current binding
-      if (strlen(argument) && !strlen(binding->argument))
+      /* append the argument to the current binding */
+      if (strlen(argument)) {
 	binding->argument = strdup(argument);
+      }
     }
   next_line:
     ;
   }
-  fclose(config);
+  fclose(f);
 
-  // add default :escape and :abort bindings to :enter'ing root bindings
+  /* parse binding arguments */
+  binding_parse_arguments(xc->root);
 
+  /* add defaults */
   for (i=0; i<self->root->num_children; i++) {
     parent = self->root->children[i];
-    if (strcmp(parent->action, ":enter") == 0) {
-      if (!binding_get_child_by_action(parent, ":escape")) {
+    if (parent->action == XC_ACTION_ENTER) {
+
+      /* create default :escape binding unless present */
+      if (!binding_get_child_by_action(parent, XC_ACTION_ESCAPE)) {
 	binding = binding_new();
-	binding->action = ":escape";
+	binding->action = XC_ACTION_ESCAPE;
 	binding->key = parent->key;
 	binding_append_child(parent, binding);
       }
-      if (!binding_get_child_by_action(parent, ":abort")) {
+
+      /* create default :abort binding unless present */
+      if (!binding_get_child_by_action(parent, XC_ACTION_ABORT)) {
 	binding = binding_new();
-	binding->action = ":abort";
+	binding->action = XC_ACTION_ABORT;
 	binding->key = key_new("C-g");
 	binding_append_child(parent, binding);
       }
@@ -291,28 +336,37 @@ void xc_parse_config(XChainKeys_t *self) {
 
   /* initialize popup window */
   self->popup = popup_new(self->display, font, fg, bg);
+  self->popup->enabled = feedback;
 
   if (self->debug) {
     printf("\n");
     printf("timeout %d\n", self->timeout);
-    printf("delay %d\n", self->delay);
-    printf("font %s\n", font);
-    printf("foreground %s\n", fg);
-    printf("background %s\n\n", bg);
+
+    if (feedback) { 
+      printf("feedback on\n");
+      printf("delay %d\n", self->delay);
+      printf("font %s\n", font);
+      printf("foreground %s\n", fg);
+      printf("background %s\n\n", bg);
+    }
+    else {
+      printf("feedback off\n\n");
+    }
     
     binding_list(self->root);
     printf("\n");
   }
+  free(buffer);
+  free(argument);
+
   free(font);
   free(fg);
   free(bg);
-  free(path);
-  free(buffer);
-  free(argument);
 }
 
 void xc_mainloop(XChainKeys_t *self) {
   Binding_t *binding;
+  Binding_t *reentry;
   XEvent event;
   KeyCode keycode;
   long now;
@@ -358,6 +412,13 @@ void xc_mainloop(XChainKeys_t *self) {
 	}
       }
     }
+  reentry:
+    if(xc->reentry != NULL) {
+      reentry = xc->reentry;
+      xc->reentry = NULL;
+      binding_activate(reentry);
+      goto reentry;
+    }
   }
 }
 
@@ -403,6 +464,32 @@ void xc_show_keys(XChainKeys_t *self) {
   }
 }
 
+void xc_send_key(XChainKeys_t *self, Key_t *key, Window window) {
+
+  XKeyEvent e;
+  char *keyspec;
+
+  if (self->debug) {
+    keyspec = key_to_str(key);
+      printf("Sending synthetic KeyPressEvent (%s) to window id %d\n", 
+	      keyspec, (int)window);
+    free(keyspec);
+  }
+
+  e.display = self->display;
+  e.window = window;
+  e.subwindow = None;
+  e.time = CurrentTime;
+  e.same_screen = True;	   
+  e.keycode = key->keycode;
+  e.state = key->modifiers;
+  e.type = KeyPress;
+  e.x = e.y = e.x_root = e.y_root = 1;
+  
+  XSendEvent(self->display, e.window, True, KeyPressMask, (XEvent *)&e);
+  XSync(self->display, False);
+}
+
 int xc_keycode_to_modmask(XChainKeys_t *self, KeyCode keycode) {
 
   int i = 0;
@@ -463,6 +550,7 @@ int main(int argc, char **argv) {
     { "help", no_argument, NULL, 'h' },
     { "version", no_argument, NULL, 'v' },
     { "keys", no_argument, NULL, 'k' },
+    { "file", no_argument, NULL, 'f' },
     { 0, 0, 0, 0 },
   };
   int option, option_index;
@@ -479,10 +567,19 @@ int main(int argc, char **argv) {
   /* parse command line arguments */
   while (1) {
 
-    option = getopt_long(argc, argv, "dhvk", options, &option_index);
+    option = getopt_long(argc, argv, "dhvkf:", options, &option_index);
     
     switch (option) {
 
+    case 'f':
+      strncpy(xc->config, optarg, strlen(optarg));
+      xc->config[strlen(optarg)] = '\0';
+      break;      
+
+    case 'k':
+      xc_show_keys(xc);
+      exit(EXIT_SUCCESS);
+
     case 'd':
       xc->debug = True;
       xc_version(xc);
@@ -498,11 +595,6 @@ int main(int argc, char **argv) {
       exit(EXIT_SUCCESS);
       break;      
 
-    case 'k':
-      xc_show_keys(xc);
-      exit(EXIT_SUCCESS);
-      break;      
-
     case -1:
       break;
 
